
#include <stdio.h>
#include "mpu_core/PointSet.h"
#include "PolygonalMesh.h"

class FileManager{

public: 
	FileManager(){}

	~FileManager(){}
	
	bool readPointFile(PointSet *ps, CString file_name){
		FILE* in = fopen(file_name, "r");
		if(in == NULL)
			return false;
		
		//read #points
		int N;
		fscanf(in, "%d", &N);

		ps->setPointSize(N);

		//read points
		for(int i=0; i<N; i++){
			float x,y,z;
			fscanf(in, "%f %f %f", &x, &y, &z);
			ps->setPoint(i, x, y, z);
		}

		//read normals
		int count = 0;
		float (*point)[3] = ps->_point;
		for(i=0; i<N; i++){
			float x,y,z;
			fscanf(in, "%f %f %f", &x, &y, &z);
			
			if(x == 0 && y == 0 && z == 0)
				continue;
			double len = sqrt(x*x + y*y + z*z);
			x /= len;
			y /= len;
			z /= len;
			
			ps->setPoint(count, point[i][0], point[i][1], point[i][2]);
			ps->setNormal(count, -x, -y, -z);
			count++;
		}

		ps->_pointN = count;

		fclose(in);
		return true;
	}

	bool writePointFile(PointSet *ps, CString file_name){
		FILE* out = fopen(file_name, "w");
		if(out == NULL)
			return false;

		int N = ps->_pointN;
		fprintf(out, "%d\n", N);

		for(int i=0; i<N; i++){
			float *p = ps->_point[i];
			fprintf(out, "%f %f %f \n", p[0], p[1], p[2]);
		}

		for(i=0; i<N; i++){
			float *p = ps->_normal[i];
			fprintf(out, "%f %f %f \n", -p[0], -p[1], -p[2]);
		}

		fclose(out);
		return true;
	}


	bool writeMeshFile(PolygonalMesh* mesh, CString file_name){
		FILE* out = fopen(file_name, "w");
		if(out == NULL)
			return false;

		int vN = mesh->vertex_N;
		int fN = mesh->face_N;
		fprintf(out, "%d\n%d\n", vN, fN);

		for(int i=0; i<vN; i++){
			float *p = mesh->vertex[i];
			fprintf(out, "%f %f %f \n", p[0], p[1], p[2]);
		}

		if(mesh->normal != NULL){
			for(i=0; i<vN; i++){
				float *p = mesh->normal[i];
				fprintf(out, "%f %f %f \n", p[0], p[1], p[2]);
			}
		}

		for(i=0; i<fN; i++){
			int pN = mesh->poly_N[i];
			int* f = mesh->face[i];
			fprintf(out, "%d ", pN);
			for(int j=0; j<pN; j++){
				fprintf(out, "%d ", f[j]);
			}
			fprintf(out, "\n");
		}

		fclose(out);
		return true;
	}
};