#ifndef RAYTRACER
#define RAYTRACER

#include "mpu_core/ImplicitOctTree.h"
#include <stdio.h>
#include <math.h>

class RayTracer{
private:
	class Color{
	public:
		float r, g, b;
		Color(){}
		Color(float r, float g, float b){
			this->r = r;
			this->g = g;
			this->b = b;
		}
		Color(Color* c){
			r = c->r;
			g = c->g;
			b = c->b;
		}
	};
public:
	Color _ambient;
	Color _back;
	float _dire[3];
	Color _direC;
	float _point[3];
	Color _pointC;

	Color _object;
	Color _object2;

	Color _objectP;
	Color _objectP2;

	Color _objectS;
	Color _objectS2;

	ImplicitPOU *func;

	int height;
	int width;

	float eye[3];
	float scale;

	//ground data
	float normal[3];
	float level;

	float ave;

	float (*trans)[3];
	int transN;
	float *error;

	RayTracer(){
		height = 400*2.0f;
		width = 275*2.0f;
		//height = 1000;
		//width = 1000;

		eye[0] = 0;
		eye[1] = 0;
		eye[2] = 40;
		scale = 0.15f/(2.75f*2.0f);
		//scale = 0.0325;
		//scale = 0.065f/2.5f;
		//scale = 0.15f;

		_ambient.r = 0.2f;
		_ambient.g = 0.2f;
		_ambient.b = 0.2f;

		_back.r = 1.0f;
		_back.g = 1.0f;
		_back.b = 1.0f;

		_dire[0] = 2/sqrt(108);
		_dire[1] = 2/sqrt(108);
		_dire[2] = 10/sqrt(108);
		
		/*
		_dire[0] = (float)(2/sqrt(68));
		_dire[1] = (float)(0/sqrt(68));
		_dire[2] = (float)(8/sqrt(68));
		*/
		_direC.r = 0.6f;
		_direC.g = 0.6f;
		_direC.b = 0.6f;

		/*
		_point[0] = 0;
		_point[1] = 15;
		_point[2] = 40;
		*/
		/*
		_point[0] = 40;
		_point[1] = 10;
		_point[2] = 25;
		*/
		_point[0] = 25;
		_point[1] = 10;
		_point[2] = 25;

		_pointC.r = 10.0f;
		_pointC.g = 10.0f;
		_pointC.b = 10.0f;

		/*
		_object.r = 0.9f;
		_object.g = 0.9f;
		_object.b = 0.9f;*/

		_object.r = 0.5f;
		_object.g = 1.0f;
		_object.b = 0.5f;

		_object2.r = 0.2f;
		_object2.g = 0.2f;
		_object2.b = 0.2f;

		normal[0] = 0;
		normal[1] = 1;
		normal[2] = 0;
		level = -10;

		_objectP.r = 0.9f;
		_objectP.g = 0.9f;
		_objectP.b = 0.9f;

		_objectP2.r = 0.0f;
		_objectP2.g = 0.0f;
		_objectP2.b = 0.0f;

		//full lucy e = 0.001
		ave = 9.5;
		//ave = 8.962690;
		//test lucy 0.001
		//ave = 8.400175f;

		transN = 5;
		trans = new float[transN][3];
		
		trans[0][0] = 0;
		trans[0][1] = 0;
		trans[0][2] = 0;

		trans[1][0] = 6;
		trans[1][1] = 0;
		trans[1][2] = -30;

		trans[2][0] = -12;
		trans[2][1] = 0;
		trans[2][2] = -60;

		trans[3][0] = 18;
		trans[3][1] = 0;
		trans[3][2] = -90;

		trans[4][0] = -24;
		trans[4][1] = 0;
		trans[4][2] = -120;
	}

	void setFunc(ImplicitPOU *func){
		this->func = func;

		error = new float[transN];
		error[0] = func->_tol;;
		error[1] = 5*error[0];
		error[2] = 5*error[1];
		error[3] = 5*error[2];
		error[4] = 5*error[3];
	}

	void createImage(CString file_name){
		Color ***image;
		image = new Color**[height];
		for(int i=0; i<height; i++){
			image[i] = new Color*[width];
			for(int j=0; j<width; j++)
				image[i][j] = new Color(&_back);
		}

		for(i=0; i<height; i++){
			for(int j=0; j<width; j++){
				float x = scale*j-0.5f*scale*width;
				float y = scale*i-0.5f*scale*height;

				float d[3] = {x-eye[0], y-eye[1], -eye[2]};

				float p[3], n[3];
				int object = -1;
				for(int k = 0; k<transN; k++){
					float screen[3];
					screen[0] = x - trans[k][0];
					screen[1] = y - trans[k][1];
					screen[2] = 0 - trans[k][2];
				
					float o[3];
					o[0] = eye[0] - trans[k][0];
					o[1] = eye[1] - trans[k][1];
					o[2] = eye[2] - trans[k][2];

					float l = (float)sqrt(d[0]*d[0]+d[1]*d[1]+d[2]*d[2]);
					d[0] /= l;
					d[1] /= l;
					d[2] /= l;

					func->_tol = error[k];
					if(func->searchZeroOnLine(p, n, o, d)){
						object = k;
						break;
					}
				}
				float p1[3];
				bool plane = ground(p1, eye, d, normal, level);
				if(object < 0 && !plane)
					continue;
				
				if(object >= 0 && plane){
					if(distFromEye(p) > distFromEye(p1)){
						object = -1;
					}
				}
			
				//render object-th translation
				if(object >= 0){
					if(object == 0 && p[0] < 0)
						convertLevelToColor(&_object, func->level(p[0], p[1], p[2]));
					else{
						_object.r = 0.5f;
						_object.g = 1.0f;
						_object.b = 0.5f;
					}

					p[0] -= 0.005f*d[0];
					p[1] -= 0.005f*d[1];
					p[2] -= 0.005f*d[2];

					n[0] = -n[0];
					n[1] = -n[1];
					n[2] = -n[2];

					p[0] += trans[object][0];
					p[1] += trans[object][1];
					p[2] += trans[object][2];

					Color* c = image[i][j];
					c->r = c->g = c->b = 0;
		
					//Directinal light
					float q[3], m[3];

					bool hit = false;
					for(int k=object; k>=0; k--){
						float p1[3] = {p[0]-trans[k][0],
							           p[1]-trans[k][1],
									   p[2]-trans[k][2]};
						func->_tol = error[k];
						hit = func->searchZeroOnLine(q, m, p1, _dire);
						if(hit)
							break;
					}

					if(!hit){
						float dot = n[0]*_dire[0] + n[1]*_dire[1] + n[2]*_dire[2];
						if(dot > 0){
							c->r += dot*_direC.r*_object.r;
							c->g += dot*_direC.g*_object.g;
							c->b += dot*_direC.b*_object.b;
						}
						if(dot > 0.1){
							float v[3];
							v[0] = 2*dot*n[0] - _dire[0];
							v[1] = 2*dot*n[1] - _dire[1];
							v[2] = 2*dot*n[2] - _dire[2];

							float h = (float)pow(v[2], 40);
							c->r += _direC.r*_object2.r*h;
							c->g += _direC.g*_object2.g*h;
							c->b += _direC.b*_object2.b*h;
						}
					}
					
					//Point light
					float v[3] = {_point[0]-p[0], 
						          _point[1]-p[1], 
								  _point[2]-p[2]};
					float l = (float)sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
					v[0] /= l;
					v[1] /= l;
					v[2] /= l;

					hit = false;
					for(k=object; k>=0; k--){
						float p1[3] = {p[0]-trans[k][0],
							           p[1]-trans[k][1],
									   p[2]-trans[k][2]};
						func->_tol = error[k];
						hit = func->searchZeroOnLine(q, m, p1, v);
						if(hit)
							break;
					}

					if(!hit){
						float dot = (n[0]*v[0] + n[1]*v[1] + n[2]*v[2]);
						if(dot > 0){
							c->r += dot*_pointC.r*_object.r/(l+1);
							c->g += dot*_pointC.g*_object.g/(l+1);
							c->b += dot*_pointC.b*_object.b/(l+1);
						}
						if(dot > 0.1){
							float v1[3];
							v1[0] = 2*dot*n[0] - v[0];
							v1[1] = 2*dot*n[1] - v[1];
							v1[2] = 2*dot*n[2] - v[2];

							float h = (float)pow(v1[2], 40);
							//float h = pow(v1[2], 10);
							c->r += _pointC.r*_object2.r*h;
							c->g += _pointC.g*_object2.g*h;
							c->b += _pointC.b*_object2.b*h;
						}
					}

					//ambient
					c->r += _ambient.r*_object.r;
					c->g += _ambient.g*_object.g;
					c->b += _ambient.b*_object.b;
					
					if(c->r > 1.0f)
						c->r = 1.0f;
					if(c->g > 1.0f)
						c->g = 1.0f;
					if(c->b > 1.0f)
						c->b = 1.0f;
				}
				else{
					p[0] = p1[0];
					p[1] = p1[1];
					p[2] = p1[2];

					n[0] = normal[0];
					n[1] = normal[1];
					n[2] = normal[2];

					Color* c = image[i][j];
					c->r = c->g = c->b = 0;
		
					//Directinal light
					float q[3], m[3];
					bool hit = false;
					for(k=0; k<transN; k++){
						float p1[3] = {p[0]-trans[k][0],
							           p[1]-trans[k][1],
									   p[2]-trans[k][2]};
						func->_tol = error[k];
						hit = func->searchZeroOnLine(q, m, p1, _dire);
						if(hit)
							break;
					}

					if(!hit){
						float dot = n[0]*_dire[0] + n[1]*_dire[1] + n[2]*_dire[2];
						if(dot > 0){
							c->r += dot*_direC.r*_objectP.r;
							c->g += dot*_direC.g*_objectP.g;
							c->b += dot*_direC.b*_objectP.b;
						}
						if(dot > 0.1){
							float v[3];
							v[0] = 2*dot*n[0] - _dire[0];
							v[1] = 2*dot*n[1] - _dire[1];
							v[2] = 2*dot*n[2] - _dire[2];

							float h = (float)pow(v[2], 40);
							c->r += _direC.r*_objectP2.r*h;
							c->g += _direC.g*_objectP2.g*h;
							c->b += _direC.b*_objectP2.b*h;
						}
					}
					
					//Point light
					float v[3] = {_point[0]-p[0], _point[1]-p[1], _point[2]-p[2]};
					float l = (float)sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
					v[0] /= l;
					v[1] /= l;
					v[2] /= l;

					hit = false;
					for(k=0; k<transN; k++){
						float p1[3] = {p[0]-trans[k][0],
							           p[1]-trans[k][1],
									   p[2]-trans[k][2]};
						func->_tol = error[k];
						hit = func->searchZeroOnLine(q, m, p1, v);
						if(hit)
							break;
					}

					if(!hit){
						float dot = (n[0]*v[0] + n[1]*v[1] + n[2]*v[2]);
						if(dot > 0){
							c->r += dot*_pointC.r*_objectP.r/(l+1);
							c->g += dot*_pointC.g*_objectP.g/(l+1);
							c->b += dot*_pointC.b*_objectP.b/(l+1);
						}
						if(dot > 0.1){
							float v1[3];
							v1[0] = 2*dot*n[0] - v[0];
							v1[1] = 2*dot*n[1] - v[1];
							v1[2] = 2*dot*n[2] - v[2];

							float h = (float)pow(v1[2], 40);
							//float h = pow(v1[2], 10);
							c->r += _pointC.r*_objectP2.r*h;
							c->g += _pointC.g*_objectP2.g*h;
							c->b += _pointC.b*_objectP2.b*h;
						}
					}

					//ambient
					c->r += _ambient.r*_objectP.r;
					c->g += _ambient.g*_objectP.g;
					c->b += _ambient.b*_objectP.b;
					
					if(c->r > 1.0f)
						c->r = 1.0f;
					if(c->g > 1.0f)
						c->g = 1.0f;
					if(c->b > 1.0f)
						c->b = 1.0f;
				}
			}
		}
		FILE* out = fopen(file_name, "w");
		fprintf(out, "P3\n");
		fprintf(out, "%d %d\n", width, height);
		fprintf(out, "255\n");
		for(i=height-1; i>-1; i--){
			for(int j=width-1; j>-1; j--){
				int r = (int)(image[i][j]->r*255);
				int g = (int)(image[i][j]->g*255);
				int b = (int)(image[i][j]->b*255);
				fprintf(out, "%d %d %d ", r, g, b);
			}
			fprintf(out, "\n");
		}
		fclose(out);
	}

	//plane defined as n[0]*x + n[1]*y + n[2]*z = l
	bool ground(float p[3], float o[3], float d[3], float n[3], float l){
		float dot = n[0]*d[0]+n[1]*d[1]+n[2]*d[2];
		if(fabs(dot) < 0.00001)
			return false;
		float t = (l-(n[0]*o[0]+n[1]*o[1]+n[2]*o[2]))/dot;
		if(t < 0)
			return false;
		p[0] = o[0] + t*d[0];
		p[1] = o[1] + t*d[1];
		p[2] = o[2] + t*d[2];
		return true;
	}

	float distFromEye(float p[3]){
		return (float)sqrt((p[0]-eye[0])*(p[0]-eye[0]) + 
			               (p[1]-eye[1])*(p[1]-eye[1]) + 
						   (p[2]-eye[2])*(p[2]-eye[2]));
	}

	void convertLevelToColor(Color* c, float level){
		c->r = c->g = c->b = 0.5;
		float v = 0.5f*(ave-level)+0.5f;

		if(v < 0){
			c->r = 1.0;
		}
		else if(v < 0.25){
			c->r=1; c->g=2.0f*v + 0.5f; 
		}
		else if(v < 0.5){
			c->r=1.0f-2.0f*v+0.5f; c->g=1;
		}
		else if(v < 0.75){
			c->g=1; c->b=2.0f*v-1.0f + 0.5;
		}
		else if(v < 1.0f){
			c->g=2.0f-2.0f*v + 0.5f; c->b=1;
		}
		else{
			c->b = 1;
		}
	}
};

#endif